package edu.berkeley.sbp.misc;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.chr.*;
import java.util.*;
import java.lang.annotation.*;
import java.lang.reflect.*;
import java.io.*;

public class Demo {
    
    public static boolean harsh = false;

    public static void main(String[] s) throws Exception {

        ReflectiveMeta m = new ReflectiveMeta();
        Tree<String> res = new CharParser(MetaGrammar.make()).parse(new FileInputStream(s[0])).expand1();
        MetaGrammar.Meta.MetaGrammarFile mgf = m.new MetaGrammarFile(res);
        MetaGrammar.BuildContext bc = new MetaGrammar.BuildContext(mgf);
        Union meta = mgf.get("s").build(bc);
        System.err.println("parsing " + s[1]);
        Tree t = new CharParser(meta).parse(new FileInputStream(s[1])).expand1();
        System.out.println("tree:\n" + t.toPrettyString());

        Reducer red = (Reducer)t.head();
        MG.Grammar g = (MG.Grammar)red.reduce(t);
        System.out.println(g);
        
        Context cx = new Context(g,m);
        Element u = null;
        for(MG.NonTerminal nt : g.nonterminals) {
            Union el = (Union)cx.get(nt.name);
            StringBuffer st = new StringBuffer();
            el.toString(st);
            System.err.println(st);
            if (nt.name.equals("s")) u = el;
        }
        System.err.println();

        System.err.println("== parsing with parsed grammar =================================================================================");
        t = new CharParser((Union)u).parse(new FileInputStream(s[1])).expand1();
        System.out.println(t.toPrettyString());

        System.err.println("== parsing with parsed-parsed grammar ==========================================================================");
        t = new CharParser(new Context(t, m).build()).parse(new FileInputStream(s[1])).expand1();
        System.out.println(t.toPrettyString());
    }

    public static class ReflectiveMeta extends MetaGrammar.Meta {
        private final Class _cl;
        private final Class[] _inner;
        public ReflectiveMeta() {
            this(MG.class,
                 new Class[] {
                     MG.Grammar.class,
                     MG.NonTerminal.class,
                     MG.AnonUn.class,
                     MG.Range.class,
                     MG.El.class,
                     MG.Seq.class,
                     MG.NonTerminalReference.class,
                     MG.StringLiteral.class,
                     MG.Tree.class,
                     MG.CharClass.class
                 });
        }
        public ReflectiveMeta(Class c, Class[] inner) {
            this._cl = c;
            this._inner = inner;
        }
        private boolean match(Method m, String s) { return match(m.getAnnotation(tag.class), null, s); }
        private boolean match(tag t, Class c, String s) {
            if (t==null) return false;
            if (t.value().equals(s)) return true;
            if (c != null && t.equals("") && c.getSimpleName().equals(s)) return true;
            return false;
        }
        private boolean match(nonterminal t, Class c, String s) {
            if (t==null) return false;
            if (t.value().equals(s)) return true;
            if (c != null && t.equals("") && c.getSimpleName().equals(s)) return true;
            return false;
        }
        private boolean match(Class c, String s, String nonTerminalName) {
            if (match((tag)c.getAnnotation(tag.class), c, s)) return true;
            if (match((nonterminal)c.getAnnotation(nonterminal.class), c, nonTerminalName)) return true;
            return false;
        }
        public boolean match(Constructor con, String s, String nonTerminalName) {
            Class c = con.getDeclaringClass();
            if (match((tag)con.getAnnotation(tag.class), null, s)) return true;
            if (match((nonterminal)con.getAnnotation(nonterminal.class), c, s)) return true;
            return false;
        }
        public Object repeatTag() {
            return new Reducer() {
                    public String toString() { return "[**]"; }
                    public Object reduce(Tree t) {
                        Object[] ret = new Object[t.numChildren()];
                        for(int i=0; i<t.numChildren(); i++) {
                            Tree tc = t.child(i);
                            if (tc.head() != null && tc.head() instanceof Reducer)
                                ret[i] = ((Reducer)tc.head()).reduce(tc);
                            else if (tc.numChildren() == 0)
                                ret[i] = tc.head();
                            else {
                                System.err.println("FIXME: don't know what to do about " + tc);
                                ret[i] = null;
                            }
                        }
                        return ret;
                    }
                };
        }
        public Sequence tryResolveTag(String tag, String nonTerminalName, Element[] els, Object[] labels, boolean[] drops) {
            Production p = new Production(tag, nonTerminalName, els, labels, drops);
            for(Method m : _cl.getMethods())
                if (new TargetMethod(m).isCompatible(p))
                    return new TargetMethod(m).makeSequence(p);
            for(Class c : _inner)
                for(Constructor con : c.getConstructors())
                    if (new TargetConstructor(con).isCompatible(p))
                        return new TargetConstructor(con).makeSequence(p);
            for(Class c : _inner)
                if (new TargetClass(c).isCompatible(p))
                    return new TargetClass(c).makeSequence(p);
            return null;
        }
        public Sequence resolveTag(String tag, String nonTerminalName, Element[] els, Object[] labels, boolean[] drops) {
            Sequence ret = tryResolveTag(tag, nonTerminalName, els, labels, drops);
            if (ret != null) return ret;
            String message = "could not find a Java method/class/ctor matching tag \""+tag+
                "\", nonterminal \""+nonTerminalName+"\" with " + els.length + " arguments";
            if (harsh) {
                throw new RuntimeException(message);
            } else {
                System.err.println(message);
                return Sequence.rewritingSequence(tag, els, labels, drops);
            }
        }
    }

    
    /**
     *  Constructors, classes, and methods with this attribute will
     *  match every production of the nonterminal called "value()"
     *  that is arg-compatible.  If value() is undefined, then the
     *  class/constructor/method name is used.
     */ 
    @Retention(RetentionPolicy.RUNTIME) public static @interface nonterminal { String value() default ""; }

    /**
     *  Constructors, classes, and methods with this attribute will
     *  match every tree tagged with "value()" that is arg-compatible.
     *  If value() is undefined, then the class/constructor/method
     *  name is used.
     */ 
    @Retention(RetentionPolicy.RUNTIME) public static @interface tag         { String value() default ""; }

    /**
     *  If any parameter to a method or field in a class has a named
     *  arg-tag, that parameter/field matches the child of the tree
     *  which either has that label or else is a reference to a
     *  nonterminal with the corresponding name.
     *  
     *  The remaining non-named arg-tags match the remaining children
     *  of the tree in sequential order.
     *
     *  If any arg-tagged parameters/fields remain, the match fails.
     *  If there were no arg-tagged parameters-fields, it is as if all
     *  of them were non-named and arg-tagged.
     *
     *  A method/constructor is arg-compatible if all of its arguments
     *  are arg-compatible.
     *
     *  A class is arg-compatible if all of its fields are
     *  arg-compatible, or if one of its constructors is arg-compatible.
     *
     */
    @Retention(RetentionPolicy.RUNTIME) public static @interface arg         { String value() default ""; }

    public static class Production {
        public String tag;
        public String nonTerminal;
        public Object[] labels;
        public boolean[] drops;
        public Element[] elements;
        public int count = 0;
        public Production(String tag, String nonTerminal, Element[] elements, Object[] labels, boolean[] drops) {
            this.tag = tag;
            this.elements = elements;
            this.nonTerminal = nonTerminal;
            this.labels = labels;
            this.drops = drops;
            for(int i=0; i<drops.length; i++)
                if (!drops[i])
                    count++;
        }
    }

    public static abstract class Target {
        public abstract String getName();
        public abstract tag getTag();
        public abstract nonterminal getNonTerminal();
        public abstract int[] buildSequence(Production p);
        public boolean isCompatible(Production p) {
            tag t = getTag();
            if (t != null &&
                (t.value().equals(p.tag) ||
                 (t.value().equals("") && p.tag.equals(getName()))))
                return buildSequence(p)!=null;

            nonterminal n = getNonTerminal();
            if (n != null &&
                (n.value().equals(p.nonTerminal) ||
                 (n.value().equals("") && getName().equals(p.nonTerminal))))
                return buildSequence(p)!=null;

            return false;
        }
        public int[] buildSequence(Production p, String[] names, arg[] argtags) {
            int argTagged = 0;
            for(int i=0; i<argtags.length; i++)
                if (argtags[i] != null)
                    argTagged++;

            // FIXME: can be smarter here
            if (names.length==p.count) {
                int[] ret = new int[p.count];
                for(int i=0; i<p.count; i++) ret[i] = i;
                return ret;
            } else if (argTagged==p.count) {
                int[] ret = new int[argtags.length];
                int j = 0;
                for(int i=0; i<argtags.length; i++)
                    ret[i] = argtags[i]==null ? -1 : (j++);
                return ret;
            } else {
                return null;
            }
        }
        public Sequence makeSequence(Production p) {
            return Sequence.rewritingSequence(new TargetReducer(p), p.elements, p.labels, p.drops);
        }
        public abstract Object plant(Object[] fields, int[] map);
        public class TargetReducer implements Reducer {
            private Production p;
            private int[] map;
            public TargetReducer(Production p) {
                this.p = p;
                this.map = buildSequence(p);
            }
            public String toString() { return "reducer-"+Target.this; }
            public Object reduce(Tree t) {
                Object[] objects = new Object[t.numChildren()];
                for(int i=0; i<t.numChildren(); i++) {
                    Tree tc = t.child(i);
                    if (tc.head() != null && tc.head() instanceof Reducer)
                        objects[i] = ((Reducer)tc.head()).reduce(tc);
                    else if (tc.numChildren() == 0)
                        objects[i] = tc.head();
                    else {
                        System.err.println("FIXME: don't know what to do about " + tc);
                        objects[i] = null;
                    }
                }
                System.err.println("input tree: " + t);
                return plant(objects, map);
            }
        }
    }

    public static interface Reducer {
        public Object reduce(Tree t);
    }

    public static class TargetClass extends Target {
        public final Class _class;
        public TargetClass(Class _class) { this._class = _class; }
        public String getName() { return _class.getSimpleName(); }
        public tag getTag() { return (tag)_class.getAnnotation(tag.class); }
        public nonterminal getNonTerminal() { return (nonterminal)_class.getAnnotation(nonterminal.class); }
        public String toString() { return _class.getSimpleName(); }
        public int[] buildSequence(Production p) {
            Field[]  f       = _class.getDeclaredFields();
            String[] names   = new String[f.length];
            arg[]    argtags = new arg[f.length];
            for(int i=0; i<f.length; i++) {
                names[i]   = f[i].getName();
                argtags[i] = f[i].getAnnotation(arg.class);
            }
            int[] ret = buildSequence(p, names, argtags);
            if (ret!=null) return ret;
            for(Constructor c : _class.getConstructors())
                if (new TargetConstructor(c).buildSequence(p)!=null)
                    return new TargetConstructor(c).buildSequence(p);
            return null;
        }
        public Object plant(Object[] fields, int[] map) {
            try {
                Object ret = _class.newInstance();
                Field[] f = _class.getFields();
                int j = 0;
                for(int i=0; i<f.length; i++)
                    if (map[i] != -1) {
                        Object tgt = Reflection.lub(fields[map[i]]);
                        if (f[i].getType() == String.class) tgt = stringify(tgt);
                        // FUGLY
                        tgt = coerce(tgt, f[i].getType());
                        System.err.println("setting a " + f[i].getType().getName() + " to " + Reflection.show(tgt));
                        f[i].set(ret, tgt);
                    }
                return ret;
            } catch (Exception e) {
                e.printStackTrace();
                throw new RuntimeException(e);
            }
        }
    }

    public static String stringify(Object o) {
        if (o==null) return "";
        if (!(o instanceof Object[])) return o.toString();
        Object[] arr = (Object[])o;
        StringBuffer ret = new StringBuffer();
        for(int i=0; i<arr.length; i++)
            ret.append(arr[i]);
        return ret.toString();
    }

    public static class TargetConstructor extends Target {
        public final Constructor _ctor;
        public TargetConstructor(Constructor _ctor) { this._ctor = _ctor; }
        public String getName() { return _ctor.getName(); }
        public tag getTag() { return (tag)_ctor.getAnnotation(tag.class); }
        public nonterminal getNonTerminal() { return (nonterminal)_ctor.getAnnotation(nonterminal.class); }
        public String toString() { return _ctor.getName(); }
        public int[] buildSequence(Production p) {
            Annotation[][] annotations = _ctor.getParameterAnnotations();
            int len = annotations.length;
            int ofs = 0;
            String name = _ctor.getDeclaringClass().getName();
            /*
            if (name.indexOf('$') > name.lastIndexOf('.')) {
                len--;
                ofs++;
            }
            */
            String[] names   = new String[len];
            arg[]    argtags = new arg[len];
            for(int i=0; i<names.length; i++)
                for(Annotation a : annotations[i+ofs])
                    if (a instanceof arg)
                        argtags[i+ofs] = (arg)a;
            return buildSequence(p, names, argtags);
        }
        public Object plant(Object[] fields, int[] map) {
            try {
                Class[] argTypes = _ctor.getParameterTypes();
                Object[] args = new Object[argTypes.length];
                int j = 0;
                for(int i=0; i<args.length; i++)
                    if (map[i] != -1) {
                        Object tgt = Reflection.lub(fields[map[i]]);
                        if (argTypes[i] == String.class) tgt = stringify(tgt);
                        // FUGLY
                        tgt = coerce(tgt, argTypes[i]);
                        System.err.println("setting a " + argTypes[i].getName() + " to " + Reflection.show(tgt));
                        args[i] = tgt;
                    }
                return _ctor.newInstance(args);
            } catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
    }
    public static class TargetMethod extends Target {
        public final Method _method;
        public TargetMethod(Method _method) { this._method = _method; }
        public String getName() { return _method.getName(); }
        public String toString() { return _method.getName(); }
        public tag getTag() { return (tag)_method.getAnnotation(tag.class); }
        public nonterminal getNonTerminal() { return (nonterminal)_method.getAnnotation(nonterminal.class); }
        public int[] buildSequence(Production p) {
            Annotation[][] annotations = _method.getParameterAnnotations();
            String[] names   = new String[annotations.length];
            arg[]    argtags = new arg[annotations.length];
            for(int i=0; i<names.length; i++)
                for(Annotation a : annotations[i])
                    if (a instanceof arg)
                        argtags[i] = (arg)a;
            int[] ret = buildSequence(p, names, argtags);
            return ret;
        }
        public Object plant(Object[] fields, int[] map) {
            try {
                Class[] argTypes = _method.getParameterTypes();
                Object[] args = new Object[argTypes.length];
                int j = 0;
                for(int i=0; i<args.length; i++)
                    if (map[i] != -1) {
                        Object tgt = Reflection.lub(fields[map[i]]);
                        if (argTypes[i] == String.class) tgt = stringify(tgt);
                        // FUGLY
                        tgt = coerce(tgt, argTypes[i]);
                        System.err.println("setting a " + argTypes[i].getName() + " to " + Reflection.show(tgt));
                        args[i] = tgt;
                    }
                System.err.println("invoking " + _method + " with " + Reflection.show(args));
                return _method.invoke(null, args);
            } catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
    }

    public static Object coerce(Object o, Class c) {
        if (o==null) return null;
        if (c.isInstance(o)) return o;
        if (c == char.class) {
            return o.toString().charAt(0);
        }
        if (c.isArray() && (c.getComponentType().isInstance(o))) {
            Object[] ret = (Object[])Array.newInstance(c.getComponentType(), 1);
            ret[0] = o;
            return ret;
        }
        if (o.getClass().isArray() && c.isArray()) {
            boolean ok = true;
            for(int i=0; i<((Object[])o).length; i++) {
                Object ob = (((Object[])o)[i]);
                if (ob != null) {
                    System.err.println("no hit with " + c.getComponentType().getName() + " on " + Reflection.show(((Object[])o)[i]));
                    ok = false;
                }
            }
            if (ok) {
                System.err.println("hit with " + c.getComponentType().getName());
                return Array.newInstance(c.getComponentType(), ((Object[])o).length);
            }
        }
        return o;
    }

    public static class MG {
        public static @tag("grammar") class Grammar {
            public NonTerminal get(String s) {
                for(NonTerminal nt : nonterminals)
                    if (nt.name.equals(s))
                        return nt;
                return null;
            }
            public @arg("NonTerminal") NonTerminal[] nonterminals;
            public String toString() {
                String ret = "[ ";
                for(NonTerminal nt : nonterminals) ret += nt + ", ";
                return ret + " ]";
            }
        }
        public abstract static class Un extends El {
            public Seq[][] sequences;
            public void build(Context cx, Union u) {
                HashSet<Sequence> bad2 = new HashSet<Sequence>();
                for(int i=0; i<sequences.length; i++) {
                    Seq[] group = sequences[i];
                    Union u2 = new Union();
                    if (sequences.length==1) u2 = u;
                    for(int j=0; j<group.length; j++) {
                        group[j].build(cx, u2, false);
                    }
                    if (sequences.length==1) break;
                    Sequence seq = Sequence.singleton(u2);
                    for(Sequence s : bad2) {
                        s.lame = true;
                        seq = seq.not(s);
                    }
                    u.add(seq);
                    bad2.add(Sequence.singleton(u2));
                }
            }
        }
        public static class NonTerminal extends Un {
            public String  name = null;
            public @nonterminal("NonTerminal") NonTerminal(@arg("Word") String name,
                                                           @arg("RHS") Seq[][] sequences) {
                this.name = name;
                this.sequences = sequences;
            }
            public Element build(Context cx) { return cx.get(name); }
        }

        public static class AnonUn extends Un {
            public @tag("(") AnonUn(Seq[][] sequences) {
                this.sequences = sequences;
            }
            public Element build(Context cx) {
                Union ret = new Union();
                build(cx, ret);
                return ret;
            }
        }

        //public static @tag void range(char c) { }
        public static class Range {
            public @tag("range") Range(char only) { first = only; last = only; }
            public @tag("-")     Range(char first, char last) { this.first = first; this.last = last; }
            public char first;
            public char last;
        }
        public static abstract class El {
            public String getLabel() { return null; }
            public String getOwnerTag() { return null; }
            public boolean drop() { return false; }
            public abstract Element build(Context cx);
        }
        public static class Drop extends El {
            public El e;
            public Drop(El e) { this.e = e; }
            public String getLabel() { return null; }
            public boolean drop() { return true; }
            public String getOwnerTag() { return e.getOwnerTag(); }
            public Element build(Context cx) { return e.build(cx); }
        }
        public static class Label extends El {
            public String label;
            public El e;
            public Label(String label, El e) { this.e = e; this.label = label; }
            public String getLabel() { return label; }
            public String getOwnerTag() { return e.getOwnerTag(); }
            public Element build(Context cx) { return e.build(cx); }
        }
        public static /*abstract*/ class Seq {
            HashSet<Seq> and = new HashSet<Seq>();
            HashSet<Seq> not = new HashSet<Seq>();
            El[] elements;
            El follow;
            String tag = null;
            boolean lame;
            public Seq(El e) { this(new El[] { e }); }
            public Seq(El[] elements) { this.elements = elements; }
            public Seq tag(String tag) { this.tag = tag; return this; }
            public Seq follow(El follow) { this.follow = follow; return this; }
            public Seq dup() {
                Seq ret = new Seq(elements);
                ret.and.addAll(and);
                ret.not.addAll(not);
                ret.follow = follow;
                ret.tag = tag;
                return ret;
            }
            public Seq and(Seq s) { and.add(s); s.lame = true; return this; }
            public Seq andnot(Seq s) { not.add(s); s.lame = true; return this; }
            public Seq separate(El sep) {
                El[] elements = new El[this.elements.length * 2 - 1];
                for(int i=0; i<this.elements.length; i++) {
                    elements[i*2]   = this.elements[i];
                    if (i<this.elements.length-1)
                        elements[i*2+1] = new Drop(sep);
                }
                this.elements = elements;
                return this;
            }
            public Sequence build(Context cx, Union u, boolean lame) {
                Sequence ret = build0(cx, lame || this.lame);
                for(Seq s : and) { Sequence dork = s.build(cx, u, true); ret = ret.and(dork); }
                for(Seq s : not) { Sequence dork = s.build(cx, u, true); ret = ret.not(dork); }
                u.add(ret);
                ret.lame = lame;
                return ret;
            }
            public Sequence build0(Context cx, boolean lame) {
                boolean unwrap = false;
                boolean dropAll = lame;
                if (tag!=null && tag.equals("[]")) unwrap  = true;
                if (tag!=null && "()".equals(tag)) dropAll = true;
                Object[] labels = new Object[elements.length];
                boolean[] drops = new boolean[elements.length];
                Element[] els = new Element[elements.length];
                for(int i=0; i<elements.length; i++) {
                    labels[i] = elements[i].getLabel();
                    drops[i]  = elements[i].drop();
                    els[i] = elements[i].build(cx);
                    if (elements[i].getOwnerTag() != null)
                        tag = elements[i].getOwnerTag();
                }
                Sequence ret = null;
                if (dropAll)     ret = Sequence.drop(els, false);
                else if (unwrap) ret = Sequence.unwrap(els, cx.rm.repeatTag(), drops);
                else if (tag!=null) {
                    ret = cx.rm.resolveTag(tag, cx.cnt, els, labels, drops);
                } else {
                    int idx = -1;
                    for(int i=0; i<els.length; i++)
                        if (!drops[i])
                            if (idx==-1) idx = i;
                            else throw new Error("multiple non-dropped elements in sequence: " + Sequence.drop(els,false));
                    if (idx != -1) ret = Sequence.singleton(els, idx);
                    else           ret = Sequence.drop(els, false);
                }
                if (this.follow != null)
                    ret.follow = MetaGrammar.infer(this.follow.build(cx));
                ret.lame = this.lame;
                return ret;
            }
        }
        public static @tag("&")   Seq  and(Seq s,         El[] elements) { return s.and(seq(elements)); }
        public static @tag("&~")  Seq  andnot(Seq s,      El[] elements) { return s.andnot(seq(elements)); }
        public static @tag("->")  Seq  arrow(Seq s, El e)                { return s.follow(e); }
        public static @tag("::")  Seq  tag(String tagname, Seq s)        { return s.tag(tagname); }
        public static @tag("/")   Seq  slash(Seq s, El e)                { return s.separate(e); }

        public static @tag("ps")  Seq  seq(El[] elements)                { return new Seq(elements); }
        public static @tag        Seq  psx(Seq s)                        { return s; }
        public static @tag(":")   El   colon(String s, El e)             { return new Label(s, e); }
        public static @tag(")")   void close(String foo)                 { throw new Error("not supported"); }
        public static @tag("()")  El   epsilon()                         { return new Constant(Union.epsilon); }

        public static @tag("nonTerminal") class NonTerminalReference extends El {
            public @arg String nonTerminal;
            public Element build(Context cx) {
                return cx.get(nonTerminal);
            }
        }
        public static class StringLiteral        extends Constant {
            public @tag("literal") StringLiteral(String string) { super(CharRange.string(string)); }
            public boolean drop() { return true; }
        }
        public static                     class CharClass            extends El {
            Range[] ranges;
            public @tag("[") CharClass(Range[] ranges) { this.ranges = ranges; }
            public Element build(Context cx) {
                edu.berkeley.sbp.util.Range.Set set = new edu.berkeley.sbp.util.Range.Set();
                for(Range r : ranges)
                        set.add(r.first, r.last);
                return CharRange.set(set);
            }
        }

        public static @tag("{")           class Tree                 extends El {
            public @arg Seq body;
            public Element build(Context cx) {
                throw new Error();
            }
        }

        public static class Rep extends El {
            public El e, sep;
            public boolean zero, many, max;
            public Rep(El e, El sep, boolean zero, boolean many, boolean max) {
                this.e = e; this.sep = sep; this.zero = zero; this.many = many; this.max = max;}
            public Element build(Context cx) {
                return (!max)
                    ? Sequence.repeat(e.build(cx),        zero, many, sep==null ? null : sep.build(cx), cx.rm.repeatTag())
                    : sep==null
                    ? Sequence.repeatMaximal(MetaGrammar.infer(e.build(cx)), zero, many,                                   cx.rm.repeatTag())
                    : Sequence.repeatMaximal(e.build(cx),                    zero, many, MetaGrammar.infer(sep.build(cx)), cx.rm.repeatTag());
            }
        }
        public static class Constant extends El {
            Element constant;
            public Constant(Element constant) { this.constant = constant; }
            public Element build(Context cx) { return constant; }
        }
        public abstract static class PostProcess extends El {
            El e;
            public PostProcess(El e) { this.e = e; }
            public Element build(Context cx) { return postProcess(e.build(cx)); }
            public abstract Element postProcess(Element e);
        }

        // FIXME: it would be nice if we could hoist this into "Rep"
        public static @tag("++")  El plusmax(final El e)                     { return new Rep(e, null, false, true, true); }
        public static @tag("+")   El plus(final El e)                        { return new Rep(e, null, false, true, false); }
        public static @tag("++/") El plusmaxfollow(final El e, final El sep) { return new Rep(e, sep,  false, true, true); }
        public static @tag("+/")  El plusfollow(final El e, final El sep)    { return new Rep(e, sep,  false, true, false); }
        public static @tag("**")  El starmax(final El e)                     { return new Rep(e, null, true,  true, true); }
        public static @tag("*")   El star(final El e)                        { return new Rep(e, null, true,  true, false); }
        public static @tag("**/") El starmaxfollow(final El e, final El sep) { return new Rep(e, sep,  true,  true, true); }
        public static @tag("*/")  El starfollow(final El e, final El sep)    { return new Rep(e, sep,  true,  true, false); }
        public static @tag("?")   El question(final El e)                    { return new Rep(e, null, true,  true, false); }

        public static @tag("!")   El bang(final El e)                        { return new Drop(e); }

        public static @tag("^")   El caret(final String s) {
            return new Drop(new Constant(CharRange.string(s)) {
                    public String getOwnerTag() { return s; }
                });
        }

        public static @tag("~")   El tilde(final El e) {
            return new PostProcess(e) {
                    public Element postProcess(Element e) {
                        return MetaGrammar.infer((Topology<Character>)Atom.toAtom(e).complement()); 
                    } }; }

        public static @tag("^^")  void doublecaret(final El e)                 { throw new Error("not implemented"); }

        //public static @tag("(")   El subexpression(Seq[][] rhs)                { return new NonTerminal(rhs); }

        public static @nonterminal("Word")    String word(String s) { return s; }
        public static @nonterminal("Quoted")  String quoted(String s) { return s; }
        public static @nonterminal("escaped") String c(char c) { return c+""; }
        public static @tag("\"\"")            String emptystring() { return ""; }
        public static @tag("\n")              String retur() { return "\n"; }
        public static @tag("\r")              String lf() { return "\r"; }

    }
    public static class Context {
        HashMap<String,Union> map = new HashMap<String,Union>();
        private MG.Grammar grammar;
        public String cnt = null;
        private ReflectiveMeta rm;
        public Context(MG.Grammar g, ReflectiveMeta rm) {
            this.grammar = g;
            this.rm = rm;
        }
        public Union build() {
            Union ret = null;
            for(MG.NonTerminal nt : grammar.nonterminals) {
                Union u = get(nt.name);
                if ("s".equals(nt.name))
                    ret = u;
            }
            return ret;
        }
        public Context(Tree t, ReflectiveMeta rm) {
            this.rm = rm;
            Reducer red = (Reducer)t.head();
            this.grammar = (MG.Grammar)red.reduce(t);
        }
        public Union peek(String name) { return map.get(name); }
        public void  put(String name, Union u) { map.put(name, u); }
        public Union get(String name) {
            Union ret = map.get(name);
            if (ret != null) return ret;
            ret = new Union(name);
            map.put(name, ret);
            MG.NonTerminal nt = grammar.get(name);
            if (nt==null) {
                System.err.println("*** warning could not find " + name);
            } else {
                String old = cnt;
                cnt = name;
                nt.build(this, ret);
                cnt = old;
            }
            return ret;
        }

    }
}
